@@ -44,16 +44,27 @@ const char *gweather_dpgettext (const char *context, const char *str) G_GNUC_FOR
 #define C_(context, str) (gweather_dpgettext (context, str))
 #define N_(str) (str)
 
+struct _GWeatherLocation {
+    char *name, *sort_name;
+    GWeatherLocation *parent, **children;
+    GWeatherLocationLevel level;
+    char *country_code, *tz_hint;
+    char *station_code, *forecast_zone, *yahoo_id, *radar;
+    double latitude, longitude;
+    gboolean latlon_valid;
+    GWeatherTimezone **zones;
+    GHashTable *metar_code_cache;
+
+    int ref_count;
+};
 
 #define WEATHER_LOCATION_CODE_LEN 4
 
-/* Why we have this in code?
-   (at least, it's not exposed in API anymore)
-*/
 typedef struct {
     gchar *name;
     gchar *code;
     gchar *zone;
+    gchar *yahoo_id;
     gchar *radar;
     gboolean latlon_valid;
     gdouble  latitude;
@@ -62,21 +73,14 @@ typedef struct {
     gchar *tz_hint;
 } WeatherLocation;
 
-WeatherLocation *       _weather_location_from_gweather_location (GWeatherLocation *gloc, const gchar *name);
+GWeatherLocation *_gweather_location_new_detached (GWeatherLocation *nearest_station,
+						   const char       *name,
+						   gboolean          latlon_valid,
+						   gdouble           latitude,
+						   gdouble           longitude);
 
-WeatherLocation *	_weather_location_new 	(const gchar *trans_name,
-						 const gchar *code,
-						 const gchar *zone,
-						 const gchar *radar,
-						 gboolean     latlon_valid,
-						 double       latitude,
-						 double       longitude,
-						 const gchar *country_code,
-						 const gchar *tz_hint);
-WeatherLocation *	_weather_location_clone	(const WeatherLocation *location);
-void			_weather_location_free	(WeatherLocation *location);
-gboolean		_weather_location_equal	(const WeatherLocation *location1,
-						 const WeatherLocation *location2);
+void              _gweather_location_update_weather_location (GWeatherLocation *gloc,
+							      WeatherLocation  *loc);
 
 /*
  * Weather information.
@@ -91,6 +95,7 @@ typedef time_t GWeatherUpdate;
 
 struct _GWeatherInfoPrivate {
     GWeatherForecastType forecast_type;
+    GWeatherProvider providers;
 
     GSettings *settings;
 
@@ -102,10 +107,11 @@ struct _GWeatherInfoPrivate {
     gboolean polarNight;
     gboolean moonValid;
     gboolean tempMinMaxValid;
-    WeatherLocation *location;
+    WeatherLocation location;
     GWeatherLocation *world;
     GWeatherLocation *glocation;
     GWeatherUpdate update;
+    GWeatherUpdate current_time;
     GWeatherSky sky;
     GWeatherConditions cond;
     GWeatherTemperature temp;
@@ -122,6 +128,7 @@ struct _GWeatherInfoPrivate {
     GWeatherMoonLatitude moonlatitude;
     gchar *forecast;
     GSList *forecast_list; /* list of GWeatherInfo* for the forecast, NULL if not available */
+    gchar *forecast_attribution;
     gchar *radar_buffer;
     gchar *radar_url;
     GdkPixbufLoader *radar_loader;
@@ -175,17 +182,17 @@ struct _GWeatherInfoPrivate {
 #define PERIGEE_LONGITUDE(d)       (282.93768193 + (d)/36525.*0.32327364)
 
 void		metar_start_open	(GWeatherInfo *info);
-void		iwin_start_open		(GWeatherInfo *info);
+gboolean	iwin_start_open		(GWeatherInfo *info);
 void		metoffice_start_open	(GWeatherInfo *info);
 void		bom_start_open		(GWeatherInfo *info);
 void		wx_start_open		(GWeatherInfo *info);
+gboolean        yahoo_start_open        (GWeatherInfo *info);
+gboolean        yrno_start_open         (GWeatherInfo *info);
 
 gboolean	metar_parse		(gchar *metar,
 					 GWeatherInfo *info);
 
-gboolean	requests_init		(GWeatherInfo *info);
-void		request_done		(GWeatherInfo *info,
-					 gboolean     ok);
+void		_gweather_info_request_done (GWeatherInfo *info);
 
 void		ecl2equ			(gdouble t,
 					 gdouble eclipLon,
@@ -193,14 +200,13 @@ void		ecl2equ			(gdouble t,
 					 gdouble *ra,
 					 gdouble *decl);
 gdouble		sunEclipLongitude	(time_t t);
-gboolean	calc_sun		(GWeatherInfo *info);
-gboolean	calc_sun_time		(GWeatherInfo *info, time_t t);
-gboolean	calc_moon		(GWeatherInfo *info);
-gboolean	calc_moon_phases	(GWeatherInfo *info, time_t *phases);
 
-void		free_forecast_list	(GWeatherInfo *info);
+void            _gweather_info_ensure_sun  (GWeatherInfo *info);
+void            _gweather_info_ensure_moon (GWeatherInfo *info);
 
-GWeatherInfo   *_gweather_info_new_clone (GWeatherInfo *info);
+void		free_forecast_list	  (GWeatherInfo *info);
+
+GWeatherInfo   *_gweather_info_new_clone  (GWeatherInfo *info);
 
 #endif /* __WEATHER_PRIV_H_ */
 
