@@ -47,13 +47,18 @@
 
 
 /**
- * SECTION:gweather-info
+ * SECTION:gweatherinfo
  * @Title: GWeatherInfo
+ *
+ * #GWeatherInfo provides a handy way to access weather conditions
+ * and forecasts from a #GWeatherLocation, aggregating multiple
+ * different web services.
+ *
+ * It includes also astronomical data such as sunrise times and
+ * moon phases.
  */
 
 
-static void _weather_internal_check (void);
-
 #define TEMPERATURE_UNIT "temperature-unit"
 #define DISTANCE_UNIT    "distance-unit"
 #define SPEED_UNIT       "speed-unit"
@@ -66,6 +71,7 @@ enum {
     PROP_WORLD,
     PROP_LOCATION,
     PROP_TYPE,
+    PROP_ENABLED_PROVIDERS,
     PROP_LAST
 };
 
@@ -107,94 +113,16 @@ gweather_dpgettext (const char *context,
     return g_dpgettext2 (GETTEXT_PACKAGE, context, str);
 }
 
-WeatherLocation *
-_weather_location_new (const gchar *name, const gchar *code,
-		       const gchar *zone, const gchar *radar,
-		       gboolean     latlon_valid,
-		       double       latitude,
-		       double       longitude,
-		       const gchar *country_code,
-		       const gchar *tz_hint)
-{
-    WeatherLocation *location;
-
-    _weather_internal_check ();
-
-    location = g_slice_new0 (WeatherLocation);
-
-    /* name and metar code must be set */
-    location->name = g_strdup (name);
-    location->code = g_strdup (code);
-
-    if (zone)
-        location->zone = g_strdup (zone);
-
-    if (radar)
-        location->radar = g_strdup (radar);
-
-    location->latlon_valid = latlon_valid;
-    location->latitude = latitude;
-    location->longitude = longitude;
-    location->country_code = g_strdup (country_code);
-    location->tz_hint = g_strdup (tz_hint);
-
-    return location;
-}
-
-WeatherLocation *
-_weather_location_clone (const WeatherLocation *location)
-{
-    WeatherLocation *clone;
-
-    g_return_val_if_fail (location != NULL, NULL);
-
-    clone = g_slice_new0 (WeatherLocation);
-
-    clone->name = g_strdup (location->name);
-    clone->country_code = g_strdup (location->country_code);
-    clone->tz_hint = g_strdup (location->tz_hint);
-
-    if (location->zone)
-	clone->zone = g_strdup (location->zone);
-
-    if (location->radar)
-	clone->radar = g_strdup (location->radar);
-
-    clone->latlon_valid = location->latlon_valid;
-    clone->latitude = location->latitude;
-    clone->longitude = location->longitude;
-
-    return clone;
-}
-
-void
+static void
 _weather_location_free (WeatherLocation *location)
 {
-    if (location) {
-        g_free (location->name);
-        g_free (location->code);
-        g_free (location->zone);
-        g_free (location->radar);
-        g_free (location->country_code);
-        g_free (location->tz_hint);
-
-        g_slice_free (WeatherLocation, location);
-    }
-}
-
-gboolean
-_weather_location_equal (const WeatherLocation *location1, const WeatherLocation *location2)
-{
-    /* if something is NULL, then it's TRUE if and only if both are NULL) */
-    if (location1 == NULL || location2 == NULL)
-        return (location1 == location2);
-    if (!location1->code || !location2->code)
-        return (location1->code == location2->code);
-    if (!location1->name || !location2->name)
-        return (location1->name == location2->name);
-
-    return ((strcmp (location1->code, location2->code) == 0) &&
-	    (strcmp (location1->name, location2->name) == 0));
+    g_free (location->name);
+    g_free (location->code);
+    g_free (location->zone);
+    g_free (location->yahoo_id);
+    g_free (location->radar);
+    g_free (location->country_code);
+    g_free (location->tz_hint);
 }
 
 static const gchar *wind_direction_str[] = {
@@ -303,10 +231,7 @@ gweather_conditions_to_string (GWeatherConditions *cond)
     }
 }
 
-/* Locals turned global to facilitate asynchronous HTTP requests */
-
-
-gboolean
+static gboolean
 requests_init (GWeatherInfo *info)
 {
     if (info->priv->requests_pending)
@@ -315,12 +240,9 @@ requests_init (GWeatherInfo *info)
     return TRUE;
 }
 
-void request_done (GWeatherInfo *info, gboolean ok)
+void
+_gweather_info_request_done (GWeatherInfo *info)
 {
-    if (ok) {
-	(void) calc_sun (info);
-	info->priv->moonValid = info->priv->valid && calc_moon (info);
-    }
     if (!--info->priv->requests_pending)
         g_signal_emit (info, gweather_info_signals[SIGNAL_UPDATED], 0);
 }
@@ -441,10 +363,12 @@ gweather_info_reset (GWeatherInfo *info)
 {
     GWeatherInfoPrivate *priv = info->priv;
 
-    if (priv->forecast)
-	g_free (priv->forecast);
+    g_free (priv->forecast);
     priv->forecast = NULL;
 
+    g_free (priv->forecast_attribution);
+    priv->forecast_attribution = NULL;
+
     free_forecast_list (info);
 
     if (priv->radar != NULL) {
@@ -453,6 +377,7 @@ gweather_info_reset (GWeatherInfo *info)
     }
 
     priv->update = 0;
+    priv->current_time = time(NULL);
     priv->sky = -1;
     priv->cond.significant = FALSE;
     priv->cond.phenomenon = GWEATHER_PHENOMENON_NONE;
@@ -499,6 +424,7 @@ void
 gweather_info_update (GWeatherInfo *info)
 {
     GWeatherInfoPrivate *priv = info->priv;
+    gboolean ok;
 
     /* Update in progress */
     if (!requests_init (info))
@@ -513,12 +439,29 @@ gweather_info_update (GWeatherInfo *info)
 #endif
     }
 
-    metar_start_open (info);
-    iwin_start_open (info);
+    if (priv->providers & GWEATHER_PROVIDER_METAR)
+	metar_start_open (info);
 
     if (priv->radar) {
         wx_start_open (info);
     }
+
+    ok = FALSE;
+    /* Try national forecast services first */
+    if (priv->providers & GWEATHER_PROVIDER_IWIN)
+	ok = iwin_start_open (info);
+    if (ok)
+	return;
+
+    /* Try Yahoo! Weather next */
+    if (priv->providers & GWEATHER_PROVIDER_YAHOO)
+	ok = yahoo_start_open (info);
+    if (ok)
+	return;
+
+    /* Try yr.no next */
+    if (priv->providers & GWEATHER_PROVIDER_YR_NO)
+	yrno_start_open (info);
 }
 
 void
@@ -542,8 +485,7 @@ gweather_info_finalize (GObject *object)
     if (priv->session)
 	g_object_unref (priv->session);
 
-    _weather_location_free (priv->location);
-    priv->location = NULL;
+    _weather_location_free (&priv->location);
 
     if (priv->glocation)
 	gweather_location_unref (priv->glocation);
@@ -593,9 +535,7 @@ gweather_info_get_location_name (GWeatherInfo *info)
 {
     g_return_val_if_fail (GWEATHER_IS_INFO (info), NULL);
 
-    if (!info->priv->location)
-	return NULL;
-    return g_strdup(info->priv->location->name);
+    return g_strdup(info->priv->location.name);
 }
 
 gchar *
@@ -929,15 +869,12 @@ gweather_info_get_sunrise (GWeatherInfo *info)
     gchar *buf;
 
     g_return_val_if_fail (GWEATHER_IS_INFO (info), NULL);
-    g_return_val_if_fail (info->priv->location, NULL);
 
     priv = info->priv;
 
-    if (!priv->location->latlon_valid)
-        return g_strdup ("-");
-    if (!priv->valid)
-        return g_strdup ("-");
-    if (!calc_sun (info))
+    _gweather_info_ensure_sun (info);
+
+    if (!priv->valid || !priv->sunriseValid)
         return g_strdup ("-");
 
     sunrise = g_date_time_new_from_unix_local (priv->sunrise);
@@ -958,15 +895,12 @@ gweather_info_get_sunset (GWeatherInfo *info)
     gchar *buf;
 
     g_return_val_if_fail (GWEATHER_IS_INFO (info), NULL);
-    g_return_val_if_fail (info->priv->location, NULL);
 
     priv = info->priv;
 
-    if (!priv->location->latlon_valid)
-        return g_strdup ("-");
-    if (!priv->valid)
-        return g_strdup ("-");
-    if (!calc_sun (info))
+    _gweather_info_ensure_sun (info);
+
+    if (!priv->valid || !priv->sunsetValid)
         return g_strdup ("-");
 
     sunset = g_date_time_new_from_unix_local (priv->sunset);
@@ -1021,6 +955,26 @@ gweather_info_get_radar (GWeatherInfo *info)
     return info->priv->radar;
 }
 
+/**
+ * gweather_info_get_attribution:
+ * @info: a #GWeatherInfo
+ *
+ * Some weather services require the application showing the
+ * data to include an attribution text, possibly including links
+ * to the service website.
+ * This must be shown prominently toghether with the data.
+ *
+ * Returns: (transfer none): the required attribution text, in Pango
+ *          markup form, or %NULL if not required
+ */
+const gchar *
+gweather_info_get_attribution (GWeatherInfo *info)
+{
+    g_return_val_if_fail (GWEATHER_IS_INFO (info), NULL);
+
+    return info->priv->forecast_attribution;
+}
+
 gchar *
 gweather_info_get_temp_summary (GWeatherInfo *info)
 {
@@ -1064,13 +1018,46 @@ gweather_info_get_weather_summary (GWeatherInfo *info)
     return out;
 }
 
+/**
+ * gweather_info_is_daytime:
+ * @info: a #GWeatherInfo
+ *
+ * Returns whether it is daytime (that is, if the sun is visible)
+ * or not at the location and the point of time referred by @info.
+ * This is mostly equivalent to comparing the return value
+ * of gweather_info_get_value_sunrise() and
+ * gweather_info_get_value_sunset(), but it accounts also
+ * for midnight sun and polar night, for locations within
+ * the Artic and Antartic circles.
+ */
+gboolean
+gweather_info_is_daytime (GWeatherInfo *info)
+{
+    GWeatherInfoPrivate *priv;
+    time_t current_time;
+
+    g_return_val_if_fail (GWEATHER_IS_INFO (info), FALSE);
+
+    priv = info->priv;
+
+    _gweather_info_ensure_sun (info);
+
+    if (priv->polarNight)
+	return FALSE;
+    if (priv->midnightSun)
+	return TRUE;
+
+    current_time = priv->current_time;
+    return ( !priv->sunriseValid || (current_time >= priv->sunrise) ) &&
+	( !priv->sunsetValid || (current_time < priv->sunset) );
+}
+
 const gchar *
 gweather_info_get_icon_name (GWeatherInfo *info)
 {
     GWeatherInfoPrivate *priv;
     GWeatherConditions   cond;
     GWeatherSky          sky;
-    time_t               current_time;
     gboolean             daytime;
     gchar*               icon;
     static gchar         icon_buffer[32];
@@ -1082,8 +1069,8 @@ gweather_info_get_icon_name (GWeatherInfo *info)
 
     priv = info->priv;
 
-    if (!priv->valid)
-        return NULL;
+    _gweather_info_ensure_sun (info);
+    _gweather_info_ensure_moon (info);
 
     cond = priv->cond;
     sky = priv->sky;
@@ -1132,17 +1119,7 @@ gweather_info_get_icon_name (GWeatherInfo *info)
         }
     }
 
-    if (priv->midnightSun ||
-	(!priv->sunriseValid && !priv->sunsetValid))
-	daytime = TRUE;
-    else if (priv->polarNight)
-	daytime = FALSE;
-    else {
-	current_time = time (NULL);
-	daytime =
-	    ( !priv->sunriseValid || (current_time >= priv->sunrise) ) &&
-	    ( !priv->sunsetValid || (current_time < priv->sunset) );
-    }
+    daytime = gweather_info_is_daytime (info);
 
     switch (sky) {
     case GWEATHER_SKY_INVALID:
@@ -1181,7 +1158,7 @@ gweather_info_get_icon_name (GWeatherInfo *info)
 	if (phase == MOON_PHASES) {
 	    phase = 0;
 	} else if (phase > 0 &&
-		   (RADIANS_TO_DEGREES(priv->location->latitude)
+		   (RADIANS_TO_DEGREES(priv->location.latitude)
 		    < moonLat)) {
 	    /*
 	     * Locations south of the moon's latitude will see the moon in the
@@ -1568,7 +1545,9 @@ gweather_info_get_value_sunrise (GWeatherInfo *info, time_t *value)
     g_return_val_if_fail (GWEATHER_IS_INFO (info), FALSE);
     g_return_val_if_fail (value != NULL, FALSE);
 
-    if (!info->priv->valid || !info->priv->sunriseValid)
+    _gweather_info_ensure_sun (info);
+
+    if (!info->priv->sunriseValid)
 	return FALSE;
 
     *value = info->priv->sunrise;
@@ -1589,7 +1568,9 @@ gweather_info_get_value_sunset (GWeatherInfo *info, time_t *value)
     g_return_val_if_fail (GWEATHER_IS_INFO (info), FALSE);
     g_return_val_if_fail (value != NULL, FALSE);
 
-    if (!info->priv->valid || !info->priv->sunsetValid)
+    _gweather_info_ensure_sun (info);
+
+    if (!info->priv->sunsetValid)
 	return FALSE;
 
     *value = info->priv->sunset;
@@ -1614,7 +1595,9 @@ gweather_info_get_value_moonphase (GWeatherInfo      *info,
     g_return_val_if_fail (value != NULL, FALSE);
     g_return_val_if_fail (lat != NULL, FALSE);
 
-    if (!info->priv->valid || !info->priv->moonValid)
+    _gweather_info_ensure_moon (info);
+
+    if (!info->priv->moonValid)
 	return FALSE;
 
     *value = info->priv->moonphase;
@@ -1703,34 +1686,6 @@ gweather_info_get_value_visibility (GWeatherInfo *info,
     return distance_value (info->priv->visibility, unit, value, info->priv->settings);
 }
 
-/**
- * weather_info_get_upcoming_moonphases:
- * @info: a #GWeatherInfo containing the time_t of interest
- * @phases: (out) (array fixed-size=4) (element-type gulong): An array of four
- *    time_t values that will hold the returned values.
- *    The values are estimates of the time of the next new, quarter, full and
- *    three-quarter moons.
- *
- * Returns: gboolean indicating success or failure
- */
-gboolean
-gweather_info_get_upcoming_moonphases (GWeatherInfo *info, time_t *phases)
-{
-    g_return_val_if_fail (GWEATHER_IS_INFO (info), FALSE);
-    g_return_val_if_fail (phases != NULL, FALSE);
-
-    return calc_moon_phases(info, phases);
-}
-
-static void
-_weather_internal_check (void)
-{
-    g_assert (G_N_ELEMENTS (wind_direction_str) == GWEATHER_WIND_LAST);
-    g_assert (G_N_ELEMENTS (sky_str) == GWEATHER_SKY_LAST);
-    g_assert (G_N_ELEMENTS (conditions_str) == GWEATHER_PHENOMENON_LAST);
-    g_assert (G_N_ELEMENTS (conditions_str[0]) == GWEATHER_QUALIFIER_LAST);
-}
-
 static void
 gweather_info_set_location_internal (GWeatherInfo     *info,
 				     GWeatherLocation *location)
@@ -1743,8 +1698,10 @@ gweather_info_set_location_internal (GWeatherInfo     *info,
 
     if (priv->glocation)
 	gweather_location_unref (priv->glocation);
-    if (priv->location)
-	_weather_location_free (priv->location);
+    _weather_location_free (&priv->location);
+
+    if (!priv->world && location)
+	priv->world = gweather_location_ref_world (location);
 
     if (!priv->world)
 	priv->world = gweather_location_new_world (FALSE);
@@ -1764,11 +1721,17 @@ gweather_info_set_location_internal (GWeatherInfo     *info,
 	priv->glocation = gweather_location_find_by_station_code (priv->world, station_code);
     }
 
-    priv->location = _weather_location_from_gweather_location (priv->glocation, name);
+    _gweather_location_update_weather_location (priv->glocation,
+						&priv->location);
+    if (name) {
+	g_free (priv->location.name);
+	priv->location.name = g_strdup (name);
+    }
 
     if (latlon_override) {
-	priv->location->latitude = DEGREES_TO_RADIANS (lat);
-	priv->location->longitude = DEGREES_TO_RADIANS (lon);
+	priv->location.latlon_valid = TRUE;
+	priv->location.latitude = DEGREES_TO_RADIANS (lat);
+	priv->location.longitude = DEGREES_TO_RADIANS (lon);
     }
 
     if (default_loc)
@@ -1792,6 +1755,32 @@ gweather_info_set_location (GWeatherInfo     *info,
     gweather_info_update (info);
 }
 
+GWeatherProvider
+gweather_info_get_enabled_providers (GWeatherInfo *info)
+{
+    g_return_val_if_fail (GWEATHER_IS_INFO (info),
+			  GWEATHER_PROVIDER_NONE);
+
+    return info->priv->providers;
+}
+
+void
+gweather_info_set_enabled_providers (GWeatherInfo     *info,
+				     GWeatherProvider  providers)
+{
+    g_return_if_fail (GWEATHER_IS_INFO (info));
+
+    if (info->priv->providers == providers)
+	return;
+
+    info->priv->providers = providers;
+
+    gweather_info_abort (info);
+    gweather_info_update (info);
+    g_object_notify (G_OBJECT (info), "enabled-providers");
+}
+
+
 static void
 gweather_info_set_property (GObject *object,
 			    guint property_id,
@@ -1811,6 +1800,37 @@ gweather_info_set_property (GObject *object,
     case PROP_TYPE:
 	priv->forecast_type = g_value_get_enum (value);
 	break;
+    case PROP_ENABLED_PROVIDERS:
+	gweather_info_set_enabled_providers (self, g_value_get_flags (value));
+	break;
+
+    default:
+	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
+    }
+}
+
+static void
+gweather_info_get_property (GObject    *object,
+			    guint       property_id,
+			    GValue     *value,
+			    GParamSpec *pspec)
+{
+    GWeatherInfo *self = GWEATHER_INFO (object);
+    GWeatherInfoPrivate *priv = self->priv;
+
+    switch (property_id) {
+    case PROP_WORLD:
+	g_value_set_boxed (value, priv->world);
+	break;
+    case PROP_LOCATION:
+	g_value_set_boxed (value, priv->glocation);
+	break;
+    case PROP_TYPE:
+	g_value_set_enum (value, priv->forecast_type);
+	break;
+    case PROP_ENABLED_PROVIDERS:
+	g_value_set_flags (value, priv->providers);
+	break;
     default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
@@ -1826,19 +1846,20 @@ gweather_info_class_init (GWeatherInfoClass *klass)
 
     gobject_class->finalize = gweather_info_finalize;
     gobject_class->set_property = gweather_info_set_property;
+    gobject_class->get_property = gweather_info_get_property;
 
     pspec = g_param_spec_boxed ("world",
 				"World",
 				"The hierarchy of locations containing the desired location",
 				GWEATHER_TYPE_LOCATION,
-				G_PARAM_STATIC_STRINGS | G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY);
+				G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
     g_object_class_install_property (gobject_class, PROP_WORLD, pspec);
 
     pspec = g_param_spec_boxed ("location",
 				"Location",
 				"The location this info represents",
 				GWEATHER_TYPE_LOCATION,
-				G_PARAM_STATIC_STRINGS | G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY);
+				G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
     g_object_class_install_property (gobject_class, PROP_LOCATION, pspec);
 
     pspec = g_param_spec_enum ("forecast-type",
@@ -1846,9 +1867,17 @@ gweather_info_class_init (GWeatherInfoClass *klass)
 			       "The type of forecast desired (list, zone or state)",
 			       GWEATHER_TYPE_FORECAST_TYPE,
 			       GWEATHER_FORECAST_LIST,
-			       G_PARAM_STATIC_STRINGS | G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY);
+			       G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
     g_object_class_install_property (gobject_class, PROP_TYPE, pspec);
 
+    pspec = g_param_spec_flags ("enabled-providers",
+				"Enabled providers",
+				"A bitmask of enabled weather service providers",
+				GWEATHER_TYPE_PROVIDER,
+				GWEATHER_PROVIDER_METAR | GWEATHER_PROVIDER_IWIN,
+				G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);
+    g_object_class_install_property (gobject_class, PROP_ENABLED_PROVIDERS, pspec);
+
     gweather_info_signals[SIGNAL_UPDATED] = g_signal_new ("updated",
 							  GWEATHER_TYPE_INFO,
 							  G_SIGNAL_RUN_FIRST,
@@ -1865,7 +1894,7 @@ gweather_info_class_init (GWeatherInfoClass *klass)
  * @forecast_type: the type of forecast requested
  *
  * Builds a new #GWeatherInfo that will provide weather information about
- * @location. The returned info will not be ready until the ::updated signal
+ * @location. The returned info will not be ready until the #GWeatherInfo::updated signal
  * is emitted.
  *
  * Returns: (transfer full): a new #GWeatherInfo
@@ -1891,7 +1920,7 @@ gweather_info_new (GWeatherLocation    *location,
  * @location: (allow-none): the desidered #GWeatherLocation (NULL for default)
  * @forecast_type: the type of forecast requested
  *
- * Similar to g_weather_info_new(), but also has a @world parameter, that allow controlling
+ * Similar to gweather_info_new(), but also has a @world parameter, that allow controlling
  * the hierarchy of #GWeatherLocation to which @location (or the default one taken from
  * GSettings) belongs.
  *
